#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Hashing et catégories
# 
# Le hashing est utilise lorsque le nombre de catégories est trop grand.

# In[1]:


get_ipython().run_line_magic('matplotlib', 'inline')


# On construit un jeu très simple avec deux catégories, une entière, une au format texte, distributé selon une loi [multinomial](https://docs.scipy.org/doc/numpy/reference/generated/numpy.random.multinomial.html#numpy.random.multinomial).

# In[2]:


import pandas
import numpy
prob = numpy.ones(100) / 100
rnd = numpy.random.binomial(100, 0.5, 10000)
df = pandas.DataFrame(dict(cat_int=rnd, cat_text=['cat%d' % i for i in rnd]))
df.head()


# In[3]:


dfc = df.drop('cat_int', axis=1).copy()
dfc['count'] = 1
gr = dfc.groupby('cat_text').sum().sort_values('count', ascending=False)
pandas.concat([gr.head(), gr.tail()])


# In[4]:


len(set(df.cat_text))


# On utilise le module [Category Encoders](http://contrib.scikit-learn.org/categorical-encoding/index.html) implémente la classe [Hashing](http://contrib.scikit-learn.org/categorical-encoding/hashing.html) qui applique une fonction de *hash* pour retourner une séquence de nombre binaires. Il y a 100 catégories distinctes. Il faut au pire 8 colonnes binaires pour représenter l'information. On vérifie avec un encoder binaire qui encode chaque catégorie de façon binaire.

# In[5]:


from category_encoders import BinaryEncoder
BinaryEncoder(cols=['cat_text']).fit_transform(df).head()


# In[6]:


from category_encoders import HashingEncoder
HashingEncoder(cols=['cat_text']).fit_transform(df).head()


# On peut réduire le nombre de colonnes. L'information est compressée mais pas de façon réversible. L'information est codée sur 4 colonnes.

# In[7]:


res = HashingEncoder(cols=['cat_text'], n_components=4, hash_method="sha256").fit_transform(df)
res.head()


# In[8]:


res['col_int'] = res.col_0 + 2 * res.col_1 + 4 * res.col_2 + 8 * res.col_3
res.head()


# In[9]:


res[['col_int', 'cat_int']].groupby('col_int').count()


# Pas tout-à-fait ce à quoi je m'attendais. On peut aussi utiliser quelque chose comme ceci : hash + encoding binaire.

# In[10]:


dfc = df.copy()
dfc['cat_hash'] = df["cat_text"].apply(lambda x: abs(hash(x)) % (2**4))
res = BinaryEncoder(cols=['cat_hash']).fit_transform(dfc)
res.head()


# In[11]:


res['col_int'] = res.cat_hash_0 + 2 * res.cat_hash_1 + 4 * res.cat_hash_2 + 8 * res.cat_hash_3
res.head()


# In[12]:


res[['col_int', 'cat_int', 'cat_text']].groupby(['col_int', 'cat_text'], as_index=False) \
        .count().pivot('cat_text', 'col_int', 'cat_int').astype(str).replace("nan", "")


# Ce qu'on espère : que deux classes sur-représentées ne soient pas encodées par la même valeur, ce qui est le cas ici. Il faudra donc augmenter la taille du hash (16 ici).

# In[13]: